/*
 * IEEE-488 driver implementation
 * Copyright (C) 1995  W. Eric Norum
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * W. Eric Norum
 * Saskatchewan Accelerator Laboratory
 * University of Saskatchewan
 * 107 North Road
 * Saskatoon, Saskatchewan, CANADA
 * S7N 5C6
 * 
 * eric@skatter.usask.ca
 */
 
#import "IEEE488.h"

#define DDM_DEBUG	1

#import <kernserv/prototypes.h>
#import <sys/errno.h>

#import <driverkit/generalFuncs.h>
#import	<driverkit/interruptMsg.h>
#import	<driverkit/kernelDriver.h>
#import <driverkit/i386/ioPorts.h>
#import <driverkit/i386/kernelDriver.h>
#import <driverkit/debugging.h>

#import "NAT4882.h"
#import "IEEE488Messages.h"
#import "IEEE488ioctl.h"

/*
 * I/O thread commands
 */
enum ioCommands {
	CMD_OPEN,
	CMD_CLOSE,
	CMD_READ,
	CMD_WRITE,
	CMD_IOCTL,
	CMD_SPOLL,
};

/*
 * Local DMA buffer
 */
#define DMA_BUFSIZE		(8*1024)	/* Pagesize? IOMallocLow contiguous? */

/*
 * Parameter keys (as listed in Default.table)
 */
#define CONTROLLER_KEY	"Controller GPIB Address"
#define TIMECODE_KEY	"Time Limit Code"

/*
 * Dynamic Debugging Module
 */
#if (defined (DDM_DEBUG))
# define DDM_INDEX	0
# define DDM_PROGRESS	0x100
# define DDM_COMMANDS	0x200
# define DDM_INTERRUPTS	0x400
# define logProgress(x,a,b,c)	IODEBUG(DDM_INDEX,DDM_PROGRESS,x,a,b,c,0,0)
# define logCommands(a)			IODEBUG(DDM_INDEX,DDM_COMMANDS,"CMD:0x%x\n",a,0,0,0,0)
# define logInterrupts(x,a,b,c,d,e)	IODEBUG(DDM_INDEX,DDM_INTERRUPTS,x,a,b,c,d,e)
#else
# define logProgress(fmt,a,b,c)
# define logCommands(a)
# define logInterrupts(x,a,b,c,d)
#endif

/*
 * Hardware port access
 */
#define PMAP(p)			(isaPorts.start+(p))
#define INB(p)			inb(PMAP(p))
#define OUTB(p,v)		outb(PMAP(p),v)
#define INB_PAGED(p)	(OUTB(NAT4882_AUXMR,NAT4882_AUXMR_PAGEIN), INB(p))
#define OUTB_PAGED(p,v)	(OUTB(NAT4882_AUXMR,NAT4882_AUXMR_PAGEIN), OUTB(p,v))

/*
 * Interthread communication (commandLock conditions)
 */
#define COMMAND_IDLE	0
#define COMMAND_IO		1
#define COMMAND_DONE	2

/*
 * Assume only one device
 */
static IEEE488 *sc;

/*
 * UNIX-style entry points
 */
static int IEEE488open (dev_t dev, int flag, int mode);
static int IEEE488close (dev_t dev, int flag, int mode);
static int IEEE488read (dev_t dev, struct uio *uio, int flag);
static int IEEE488write (dev_t dev, struct uio *uio, int flag);
static int IEEE488ioctl (dev_t dev, int cmd, caddr_t data, int flag);

@implementation IEEE488

/*
 * Set up NAT4882 hardware
 */
- (BOOL)init4882
{
	int ksr;

	/*
	 * Verify that device exists and is a GPIB-PCII/IIA
	 */
	ksr = INB_PAGED (NAT4882_KSR_PAGED);
	if (((ksr & NAT4882_KSR_MODE) == 0)
	 || ((ksr & NAT4882_KSR_VMASK) < 0x10)) {
		IOLog ("%s: Device is not a GPIB-PCIIA in 7210 mode.  KSR:0x%x\n", [self name], ksr);
		return NO;
	}
	IOLog ("%s: GPIB-PCIIA chip revision %d.\n", [self name], ksr >> 4);
	return YES;
}

/*
 ************************************************************************
 *							DriverKit methods							*
 ************************************************************************
 */

/*
 * Probe, configure board and init new instance.
 */
+ (BOOL)probe:description
{
	IEEE488 *			dev;
	BOOL				ret;

	/*
	 * Allocate an instance
	 */
	dev = [self alloc];
	if (dev == nil) {
		IOLog ("%s: Failed to allocate instance\n", [self name]) ;
		return NO;
	}

	/*
	 * Initialize the instance
	 */
	if ([dev initFromDeviceDescription:description] == nil) {
		IOLog ("%s: Failed to init Device Description\n", [self name]);
		[dev free];
		return NO;
	}

	/*
	 * Link the instance into the kernel
	 */
	ret = [self addToCdevswFromDescription:description
										open:  (IOSwitchFunc)IEEE488open	
										close: (IOSwitchFunc)IEEE488close
										read:  (IOSwitchFunc)IEEE488read
										write: (IOSwitchFunc)IEEE488write
										ioctl: (IOSwitchFunc)IEEE488ioctl
										stop:  (IOSwitchFunc)nulldev
										reset: (IOSwitchFunc)nulldev
										select:(IOSwitchFunc)nodev
										mmap:  (IOSwitchFunc)nodev
										getc:  (IOSwitchFunc)nodev
										putc:  (IOSwitchFunc)nodev];
	if (ret != YES) {
		IOLog ("%s: addToCdevswFromDescription failed.\n", [self name]);
		[dev free];
		return NO;
    }
	IOLog ("%s: addToCdevswFromDescription major %d\n", [self name], [self characterMajor]);
	return YES;
}

- (BOOL)getInt:(unsigned int *)ip fromConfig:config stringKey:(const char *)key
{
	const char *value, *cp;

	if ((value = [config valueForStringKey:key]) == NULL) {
		IOLog ("%s: No value for key `%s'\n", [self name], key);
		return NO;
	}
	*ip = 0;
	for (cp = value ; *cp != '\0' ; cp++) {
		if ((*cp >= '0') && (*cp <= '9')) {
			*ip = *ip * 10 + *cp - '0';
		}
		else {
			IOLog ("%s: Bad value (%s) for key `%s'\n", [self name], value, key);
			[config freeString:value];
			return NO;
		}
	}
	[config freeString:value];
	return YES;
}

/*
 * Initialize the new instance.
 */
- initFromDeviceDescription:deviceDescription
{
	IOConfigTable *config = [deviceDescription configTable];
	IOReturn r;
	char cbuf[40];

    if ([super initFromDeviceDescription:deviceDescription] == nil)
    	return nil;

	/*
	 * Verify that the description is reasonable
	 */
	if (([deviceDescription numInterrupts] != 1)
	 || ([deviceDescription numChannels] != 1)
	 || ([deviceDescription numMemoryRanges] != 0)
	 || ([deviceDescription numPortRanges] != 1)) {
		IOLog ("%s: Bad device description!\n", [[self class] name]);
		return nil;
	}
	isaPorts = *[deviceDescription portRangeList];
	if (isaPorts.size != 8) {
		IOLog ("%s: Bad number of ports in device description!\n", [[self class] name]);
		return nil;
	}
	if ([deviceDescription interrupt] == 2) {
		int	aList[1];
		aList[0] = 9;
		[deviceDescription setInterruptList:aList num:1];
	}

    [self setName:[config valueForStringKey:"Driver Name"]];
    [self setUnit:0];
	sprintf (cbuf, "%x-%x", isaPorts.start, isaPorts.start+isaPorts.size-1);
    [self setLocation:cbuf];
    IOLog ("%s: Interrupt:%d Channel:%d\n",	[self name],
											[deviceDescription interrupt], 
											[deviceDescription channel]);

	if (![self getInt:&myAddress fromConfig:config stringKey:CONTROLLER_KEY])
		return nil;
	if (![self getInt:&timeLimitCode fromConfig:config stringKey:TIMECODE_KEY])
		return nil;
	IOLog ("%s: Controller:%d  Timeout code:0x%x\n", [self name], myAddress, timeLimitCode);

	/*
	 * Start the I/O thread
	 */
    r = [self startIOThread];
    if (r != IO_R_SUCCESS)
		IOLog ("%s: Can't start I/O thread: %s\n", [self name], [self stringFromReturn:r]);

	/*
	 * Set up inter-thread communication
	 */
	commandLock = [[NXConditionLock alloc] initWith:COMMAND_IDLE];
	commandMessage.msg_id = IO_COMMAND_MSG;
	commandMessage.msg_size = sizeof commandMessage;
	commandMessage.msg_type = MSG_TYPE_NORMAL;
	commandMessage.msg_remote_port = IOConvertPort ([self interruptPort],
													IO_KernelIOTask, IO_Kernel);
	commandMessage.msg_local_port = PORT_NULL;
	sc = self;

	/*
	 * Set up hardware resources
	 */
	if ([self isEISAPresent]) {
		r = [self setDMATransferWidth:IO_8Bit forChannel:0];
		if (r != IO_R_SUCCESS) {
			IOLog ("%s: Can't set DMA transfer width:%s\n", [self name], [self stringFromReturn:r]);
			return nil;
		}
		r = [self setDMATiming:IO_TypeA forChannel:0];
		if (r != IO_R_SUCCESS) {
			IOLog ("%s: Can't set DMA timing :%s\n", [self name], [self stringFromReturn:r]);
			return nil;
		}
		r = [self setEOPAsOutput:YES forChannel:0];
		if (r != IO_R_SUCCESS) {
			IOLog ("%s: Can't set DMA EOP direction :%s\n", [self name], [self stringFromReturn:r]);
			return nil;
		}
	}
	r = [self setTransferMode:IO_Single forChannel:0];
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't set DMA transfer mode:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
	dmaBuffer = IOMallocLow (DMA_BUFSIZE);
	if (dmaBuffer == NULL) {
		IOLog ("Can't IOMallocLow DMA buffer\n");
		return nil;
	}
	r = IOPhysicalFromVirtual (IOVmTaskSelf(), (vm_address_t)dmaBuffer, &dmaBufferPhysicalAddress);
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't get DMA buffer physical address:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
    [self registerDevice];

	/*
	 * Set up the hardware
	 */
	if ([self init4882] == NO)
		return nil;
	r = [self enableInterrupt:0];
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't enable interrupt:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
	r = [self enableChannel:0];
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't enable channel:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
    return self; 
}

/*
 * Pass values to user-level routines
 */
- (IOReturn)getIntValues:(unsigned int *)values
				forParameter:(IOParameterName)parameter
				count:(unsigned int *)count
{
	if ((strcmp (parameter, "IOMajorDevice") == 0) && (*count == 1)) {
		values[0] = [[self class] characterMajor];
		return IO_R_SUCCESS;
	}
	else if ((strcmp (parameter, CONTROLLER_KEY) == 0) && (*count == 1)) {
		values[0] = myAddress;
		return IO_R_SUCCESS;
	}
	else if ((strcmp (parameter, TIMECODE_KEY) == 0) && (*count == 1)) {
		values[0] = timeLimitCode;
		return IO_R_SUCCESS;
	}
	return [super getIntValues:values forParameter:parameter count:count];
}

/*
 * Get values from user-level routines
 */
- (IOReturn)setIntValues:(unsigned int *)values
				forParameter:(IOParameterName)parameter
				count:(unsigned int)count
{
	if ((strcmp (parameter, TIMECODE_KEY) == 0) && (count == 1)) {
		if ((values[0] >= 12) && (values[0] <= 15)) {
			timeLimitCode = values[0];
			return IO_R_SUCCESS;
		}
		else {
			return IO_R_INVALID_ARG;
		}
	}
	return [super setIntValues:values forParameter:parameter count:count];
}

/*
 ************************************************************************
 *							I/O Thread Methods							*
 ************************************************************************
 */
- (void)finishClosing
{
	logProgress ("Cease GPIB operation\n", 0, 0, 0);
	OUTB (NAT4882_IMR2, 0);			/* Disable interrupts */
	OUTB (NAT4882_IMR1, 0);
	OUTB_PAGED (NAT4882_IMR0_PAGED, 0);
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_CREN);	/* Clear remote enable */
}

- (void)interfaceClear
{
	logProgress ("IFC\n", 0, 0, 0);
	lastTalker = lastListener = -1;
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_SIFC);	/* Set IFC */
	IODelay (100);								/* Wait for 100 usec */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_CIFC);	/* Clear IFC */
}

- (void)forceLocal
{
	logProgress ("Force LOCAL\n", 0, 0, 0);
	lastTalker = lastListener = -1;
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_CREN);	/* Clear REN */
	IODelay (100);								/* Wait for 100 usec */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_SREN);	/* Set REN */
}

- (void)initChip
{
	logProgress ("Initialize With ports 0x%x (%d)\n", isaPorts.start, isaPorts.size, 0);
	softCOflag = NO;
	softSRQIflag = NO;
	finishHandshake = NO;
	ioState = IOS_IDLE;

	/*
	 * Reset chip
	 */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_CR);

	/*
	 * Allow high-speed transfers
	 */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_B(NAT4882_AUXRB_TRI));

	/*
	 * Don't talk if there are no listeners, don't hang onto END
	 */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_G(NAT4882_AUXRG_NTNL|NAT4882_AUXRG_CHES));

	/*
	 * Set interface function to T/L, normal pullup enable
	 */
	OUTB (NAT4882_ADMR, NAT4882_ADMR_TRM1|NAT4882_ADMR_TRM0|NAT4882_ADMR_ADM0);
	OUTB (NAT4882_ADR, myAddress);
	OUTB (NAT4882_ADR, NAT4882_ADR_ARS|NAT4882_ADR_DT|NAT4882_ADR_DL);

	/*
	 * Enable interrupts
	 */
	OUTB (NAT4882_IMR1, NAT4882_IMR1_ERR_IE);
	OUTB (NAT4882_IMR2, NAT4882_IMR2_SRQI_IE|NAT4882_IMR2_CO_IE);
	OUTB_PAGED (NAT4882_IMR0_PAGED, NAT4882_IMR0_GLINT|NAT4882_IMR0_TO_IE);

	/*
	 * Set chip to power-on state
	 */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_PON);

	/*
	 * Clear interface
	 */
	[self interfaceClear];

	/*
	 * Set remote enable
	 */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_SREN);
}

- (void)completeCommand:(int)errno
{
	/*
	 * Disable timer
	 */
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_J (0));
	logProgress ("Complete, errno:%d state:%d flags:0x%x\n", errno, ioState, openFlags);
	if (errno) {
		if (ioEISADmaBuffer) {
			logProgress ("Terminate with DMA count:%d\n", [self currentCountForChannel:0], 0, 0);
			[self abortDMABuffer:ioEISADmaBuffer];
			ioEISADmaBuffer = NULL;
		}
		[self initChip];
	}
	OUTB (NAT4882_IMR1, NAT4882_IMR1_ERR_IE);
	OUTB (NAT4882_IMR2, NAT4882_IMR2_SRQI_IE|NAT4882_IMR2_CO_IE);
	OUTB_PAGED (NAT4882_IMR0_PAGED, NAT4882_IMR0_GLINT|NAT4882_IMR0_TO_IE);
	ioState = IOS_IDLE;
	commandErrno = errno;
	if (openFlags == 0)
		[self finishClosing];
	[commandLock unlockWith:COMMAND_DONE];
}

- (void)sendNextCommandByte
{
	softCOflag = NO;
	if (ieee488CommandCount) {
		logCommands (ieee488Commands[ieee488CommandIndex]);
		OUTB (NAT4882_CDOR, ieee488Commands[ieee488CommandIndex++]);
	}
	else {
		logProgress ("EPP\n", 0, 0, 0);
		ioState = IOS_PPOLL;
		OUTB (NAT4882_AUXMR, NAT4882_AUXMR_EPP);
	}
}

/*
 * Send a sequence of commands.  Operation depends on value of first argument:
 *	1) count > 0
 *		Subsequent `count' arguments are the multiline messages
 *		to be placed in ieee488Commands and transmitted.
 *	2) count == 0
 *		Conduct a parallel poll.
 *	3) count < 0
 *		ieee488Commands and ieee488CommandCount are already set up.
 */
- (void)sendCommands:(int)count, ...
{
	va_list ap;

	va_start (ap, cmd);
	if (INB (NAT4882_ADSR) & NAT4882_ADSR_ATNBAR) {
		logProgress ("TCA\n", 0, 0, 0);
		OUTB (NAT4882_AUXMR, NAT4882_AUXMR_TCA);
		finishHandshake = NO;
	}
	ioState = IOS_COMMAND;
	ieee488CommandIndex = 0;
	if (count > 0) {
		ieee488CommandCount = 0;
		while (count--)
			ieee488Commands[ieee488CommandCount++] = va_arg (ap, int);
	}
	else if (count == 0) {
		ieee488CommandCount = 0;
	}
	if (softCOflag == YES)
		[self sendNextCommandByte];
}

- (int)startDMA
{
	IOReturn r;
	unsigned int phys = dmaBufferPhysicalAddress;
	int adsr;

	ioState = IOS_DATA;
	if (uio->uio_resid <= DMA_BUFSIZE)
		dmaLength = uio->uio_resid;
	else
		dmaLength = DMA_BUFSIZE;
	ioEISADmaBuffer = [self createDMABufferFor:&phys
											length:dmaLength
											read:commandType == CMD_READ
											needsLowMemory:YES
											limitSize:YES];
	logProgress ("Phys:0x%x  phys:0x%x length:%d\n", dmaBufferPhysicalAddress, phys, dmaLength);
	if (ioEISADmaBuffer == NULL)
		return ENOSPC;
	adsr = INB (NAT4882_ADSR);
	if ((commandType == CMD_READ)
	 || ((commandType == CMD_WRITE) && (dmaLength == uio->uio_resid))) {
		logProgress ("ACC, DMA I/O, Clear SYNC\n", 0, 0, 0);
		OUTB (NAT4882_AUXMR, NAT4882_AUXMR_I (NAT4882_AUXRI_ACC));
	}
	else {
		logProgress ("No ACC, DMA I/O, Clear SYNC\n", 0, 0, 0);
		OUTB (NAT4882_AUXMR, NAT4882_AUXMR_I (0));
	}
	if (commandType == CMD_READ)
		OUTB (NAT4882_IMR2, NAT4882_IMR2_SRQI_IE|NAT4882_IMR2_DMAI);
	else
		OUTB (NAT4882_IMR2, NAT4882_IMR2_SRQI_IE|NAT4882_IMR2_DMAO);
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_CLRSYNC);
	OUTB_PAGED (NAT4882_IMR0_PAGED, NAT4882_IMR0_GLINT|NAT4882_IMR0_TO_IE|NAT4882_IMR0_SYNC_IE);
	r = [self startDMAForBuffer:ioEISADmaBuffer channel:0];
    if (r != IO_R_SUCCESS) {
		[self abortDMABuffer:ioEISADmaBuffer];
		IOLog ("%s: Can't start DMA: %s\n", [self name], [self stringFromReturn:r]);
		ioEISADmaBuffer = NULL;
		return EIO;
	}
	if ((adsr & NAT4882_ADSR_ATNBAR) == 0) {
		if (commandType == CMD_READ) {
			logProgress ("ADS:0x%x TCSE GTS\n", adsr, 0, 0);
			OUTB (NAT4882_AUXMR, NAT4882_AUXMR_TCSE);
		}
		else {
			logProgress ("ADS:0x%x GTS\n", adsr, 0, 0);
		}
		OUTB (NAT4882_AUXMR, NAT4882_AUXMR_GTS);
		softCOflag = NO;
	}
	if (finishHandshake) {
		logProgress ("FHS\n", 0, 0, 0);
		OUTB (NAT4882_AUXMR, NAT4882_AUXMR_FHS);
		finishHandshake = NO;
	}
	OUTB (NAT4882_IMR1, NAT4882_IMR1_ENDRX_IE|NAT4882_IMR1_ERR_IE);
	return 0;
}

- (void)interruptOccurred
{
	int isr0, isr1, isr2;
	unsigned int count;
	int err;

	for (;;) {
		isr2 = INB (NAT4882_ISR2);
		if ((isr2 & NAT4882_ISR2_INT) == 0)
			break;
		isr1 = INB (NAT4882_ISR1);
		isr0 = INB_PAGED (NAT4882_ISR0_PAGED);
		logInterrupts ("ISR0:0x%x ISR1:0x%x ISR2:0x%x ADSR:0x%x ADR1:0x%x\n",
															isr0, isr1, isr2,
															INB (NAT4882_ADSR),
															INB (NAT4882_ADR1));
		logInterrupts ("State:%d BSR:0x%x\n", ioState,
												INB_PAGED (NAT4882_BSR_PAGED),
												0, 0, 0);
		if (isr2 & NAT4882_ISR2_SRQI)
			softSRQIflag = YES;
		if ((isr1 & NAT4882_ISR1_ERR) && (ioState != IOS_IDLE)) {
			[self completeCommand:ENODEV];
			return;
		}
		switch (ioState) {
		case IOS_IDLE:
		case IOS_COMMAND:
		case IOS_PPOLL:
			break;

		case IOS_SPOLL:
			if (isr1 & NAT4882_ISR1_DI) {
				*(int *)ioctlData = INB (NAT4882_DIR);
				logProgress ("SPOLL:0x%x\n", *(int *)ioctlData, 0, 0);
				[self sendCommands:1, SPD_488];
			}
			break;

		case IOS_DATA:
			if ((isr0 & NAT4882_ISR0_SYNC) || (isr1 & NAT4882_ISR1_END)) {
				unsigned int ndone;
				count = [self currentCountForChannel:0];
				ndone = dmaLength - count;
				uio->uio_resid -= ndone;
				logProgress ("Free DMA after %d bytes, %d left.\n",
													ndone, uio->uio_resid, 0);
				[self freeDMABuffer:ioEISADmaBuffer];
				ioEISADmaBuffer = NULL;
				if (commandType == CMD_READ) {
					if (isr1 & NAT4882_ISR1_END) {
						readFinished |= (1 << target);
						logProgress ("FHS\n", 0, 0, 0);
						OUTB (NAT4882_AUXMR, NAT4882_AUXMR_FHS);
						finishHandshake = NO;
					}
					else {
						finishHandshake = YES;
					}
				}
				else {
					if (uio->uio_resid == 0) {
						logProgress ("TCA\n", 0, 0, 0);
						OUTB (NAT4882_AUXMR, NAT4882_AUXMR_TCA);
					}
				}
				[self completeCommand:0];
			}
			break;
		}
		if (isr2 & NAT4882_ISR2_CO) {
			switch (ioState) {
			case IOS_COMMAND:
				if (ieee488CommandCount && (ieee488CommandIndex == ieee488CommandCount)) {
					softCOflag = YES;
					switch (commandType) {
					case CMD_SPOLL:
						if (ieee488Commands[0] == SPD_488) {
							[self completeCommand:0];
						}
						else {
							/*
							 * Switch to interrupt-driven input
							 */
							OUTB (NAT4882_IMR1, NAT4882_IMR1_ERR_IE|NAT4882_IMR1_DI_IE);
							logProgress ("LTN TCS GTS\n", 0, 0, 0);
							OUTB (NAT4882_AUXMR, NAT4882_AUXMR_LTN);
							OUTB (NAT4882_AUXMR, NAT4882_AUXMR_TCS);
							OUTB (NAT4882_AUXMR, NAT4882_AUXMR_GTS);
							softCOflag = NO;
							ioState = IOS_SPOLL;
						}
						break;

					case CMD_IOCTL:
						[self completeCommand:0];
						break;

					case CMD_CLOSE:
						[self completeCommand:0];
						break;

					case CMD_READ:
					case CMD_WRITE:
						if ((err = [self startDMA]) != 0)
							[self completeCommand:err];
						break;
					}
				}
				else {
					[self sendNextCommandByte];
				}
				break;

			case IOS_PPOLL:
				softCOflag = YES;
				*(int *)ioctlData = INB (NAT4882_CPTR);
				logProgress ("PPOLL:0x%x\n", *(int *)ioctlData, 0, 0);
				[self completeCommand:0];
				break;

			default:
				softCOflag = YES;
				break;
			}
		}
		if ((isr0 & NAT4882_ISR0_TO) && (ioState != IOS_IDLE)) {
			[self completeCommand:ETIMEDOUT];
			return;
		}
	}
}

- (int)startRead
{
	int err;

	readFinished &= ~(1 << target);
	if ((lastTalker != target) || (lastListener != myAddress)) {
		lastTalker = target;
		lastListener = myAddress;
		logProgress ("Send commands for read.\n", 0, 0, 0);
		[self sendCommands:4, UNT_488,
								UNL_488,
								TAG_488 + target,
								LAG_488 + myAddress];
	}
	else if ((err = [self startDMA]) != 0) {
		return err;
	}
	return 0;
}

- (int)startWrite
{
	int err;

	if ((lastTalker != myAddress) || (lastListener != target)) {
		lastTalker = myAddress;
		lastListener = target;
		logProgress ("Send commands for write.\n", 0, 0, 0);
		[self sendCommands:4, UNT_488,
								UNL_488,
								TAG_488 + myAddress,
								LAG_488 + target];
	}
	else if ((err = [self startDMA]) != 0) {
		return err;
	}
	return 0;
}

- (void)makeListenerList:(int)listeners
{
	int bit, i;

	bit = 0x1;
	i = 0;
	while ((listeners != 0) && (i <= 31)) {
		if (listeners & bit) {
			listeners &= ~bit;
			if (i != myAddress)
				ieee488Commands[ieee488CommandCount++] = LAG_488 + i;
		}
		i++;
		bit <<= 1;
	}
}

- (void)startIEEE488Operation
{
	lastTalker = lastListener = -1;
	switch (ioctlCmd) {
	case IEEE488_GTL:
		[self sendCommands:3, UNL_488, LAG_488+target, GTL_488];
		break;

	case IEEE488_SDC:
		[self sendCommands:3, UNL_488, LAG_488+target, SDC_488];
		break;

	case IEEE488_GET:
		[self sendCommands:3, UNL_488, LAG_488+target, GET_488];
		break;

	case IEEE488_LLO:
		[self sendCommands:1, LLO_488];
		break;

	case IEEE488_DCL:
		[self sendCommands:1, DCL_488];
		break;

	case IEEE488_SPOLL:
		commandType = CMD_SPOLL;
		[self sendCommands:4, UNT_488, UNL_488, SPE_488, TAG_488+target];
		break;

	case IEEE488_PPU:
		[self sendCommands:1, PPU_488];
		break;

	case IEEE488_PPD:
		[self sendCommands:3, UNL_488, LAG_488+target, PPD_488];
		break;

	case IEEE488_PPC:
		[self sendCommands:5, UNL_488, LAG_488+target, PPC_488,
									PPE_488+(*(int *)ioctlData & 0xF), UNL_488];
		break;

	case IEEE488_PPOLL:
		[self sendCommands:0];
		break;

	case IEEE488_GETLIST:
		ieee488CommandCount = 0;
		ieee488Commands[ieee488CommandCount++] = UNL_488;
		[self makeListenerList:*(int *)ioctlData];
		ieee488Commands[ieee488CommandCount++] = GET_488;
		ieee488Commands[ieee488CommandCount++] = UNL_488;
		[self sendCommands:-1];
		break;
	}
}

- (int)startIoctl
{
	switch (ioctlCmd) {
	case IEEE488_GTL:
	case IEEE488_SDC:
	case IEEE488_GET:
	case IEEE488_LLO:
	case IEEE488_DCL:
	case IEEE488_SPOLL:
	case IEEE488_PPU:
	case IEEE488_PPD:
	case IEEE488_PPC:
	case IEEE488_PPOLL:
	case IEEE488_GETLIST:
		[self startIEEE488Operation];
		break;

	default:
		return EINVAL;
	}
	return 0;
}

- (void)commandRequestOccurred
{
	int err = 0;

	[commandLock lockWhen:COMMAND_IO];
	logProgress ("Start timer -- code 0x%x\n", timeLimitCode, 0, 0);
	OUTB (NAT4882_AUXMR, NAT4882_AUXMR_J (timeLimitCode));
	switch (commandType) {
	case CMD_OPEN:
		if (openFlags == 0)
			[self initChip];
		openFlags |= 1 << target;
		break;

	case CMD_CLOSE:
		openFlags &= ~(1 << target);
		lastTalker = lastListener = -1;
		[self sendCommands:4, UNL_488,
									TAG_488 + myAddress,
									LAG_488 + target,
									GTL_488];
		return;

	case CMD_READ:
		if ((err = [self startRead]) == 0)
			return;
		break;

	case CMD_WRITE:
		if ((err = [self startWrite]) == 0)
			return;
		break;

	case CMD_IOCTL:
		switch (ioctlCmd) {
		case IEEE488_IFC:
			[self interfaceClear];
			break;
		
		case IEEE488_LOCAL:
			[self forceLocal];
			break;
		
		case IEEE488_GETSRQ:
			*(int *)ioctlData = softSRQIflag;
			softSRQIflag = NO;
			break;
		
		default:
			if ((err = [self startIoctl]) == 0)
				return;
			break;
		}
		break;

	default:
		err = EINVAL;
		break;
	}
	[self completeCommand:err];
}

/*
 ************************************************************************
 *	Interface between UNIX-style entry points and DriverKit methods		*
 ************************************************************************
 */
- (int)IEEE488ExecuteCommand:(char)cmd forDevice:(int)device
{
	kern_return_t r;

	/*
	 * Set up information common to all commands
	 */
	target = device;
	commandType = cmd;

	/*
	 * Tell the I/O thread about the new command
	 */
	[commandLock unlockWith:COMMAND_IO];
	r = msg_send_from_kernel (&commandMessage, MSG_OPTION_NONE, 0);
	if (r != SEND_SUCCESS) {
		IOLog ("Can't send message to I/O thread: %d\n", r);
		return EIO;
	} 

	/*
	 * Await completion
	 */
	[commandLock lockWhen:COMMAND_DONE];

	/*
	 * Return the results
	 */
	return commandErrno;
}

#define logReg(x,a,b) logProgress(x,a,b,0)
- (void)showReg
{
logReg ("Registers\n", 0, 0);
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_DIR), INB(NAT4882_DIR));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_ISR1), INB(NAT4882_ISR1));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_ISR2), INB(NAT4882_ISR2));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_SPSR), INB(NAT4882_SPSR));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_ADSR), INB(NAT4882_ADSR));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_CPTR), INB(NAT4882_CPTR));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_ADR0), INB(NAT4882_ADR0));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_ADR1), INB(NAT4882_ADR1));
logReg ("Paged registers\n", 0, 0);
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_KSR_PAGED), INB_PAGED(NAT4882_KSR_PAGED));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_SASR_PAGED), INB_PAGED(NAT4882_SASR_PAGED));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_ISR0_PAGED), INB_PAGED(NAT4882_ISR0_PAGED));
logReg ("0x%x: 0x%x\n", PMAP(NAT4882_BSR_PAGED), INB_PAGED(NAT4882_BSR_PAGED));
}
@end

/*
 ************************************************************************
 *			Good old-fashioned UNIX driver entry points					*
 ************************************************************************
 */
static int
IEEE488open (dev_t dev, int flag, int mode)
{	
	int err;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	err = [sc IEEE488ExecuteCommand:CMD_OPEN forDevice:minor(dev)];
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
IEEE488close (dev_t dev, int flag, int mode)
{
	int err;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	err = [sc IEEE488ExecuteCommand:CMD_CLOSE forDevice:minor(dev)];
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
IEEE488read (dev_t dev, struct uio *uio, int flag)
{
	struct iovec *iov = uio->uio_iov;
	int ndone, nleft, ncopy;
	int err = 0;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	sc->uio = uio;
	nleft = ndone = 0;
	while (uio->uio_resid) {
		if (nleft == 0) {
			nleft = uio->uio_resid;
			err = [sc IEEE488ExecuteCommand:CMD_READ forDevice:minor(dev)];
			if (err)
				break;
			ndone = 0;
			nleft -= uio->uio_resid;
		}
		if (iov->iov_len > nleft)
			ncopy = nleft;
		else
			ncopy = iov->iov_len;
		if (copyout (sc->dmaBuffer + ndone, iov->iov_base, ncopy)) {
			err = EFAULT;
			break;
		}
		ndone += ncopy;
		nleft -= ncopy;
		iov->iov_base += ncopy;
		iov->iov_len -= ncopy;
		if (iov->iov_len == 0) {
			uio->uio_iovcnt--;
			iov++;
		}
		if (sc->readFinished & (1 << minor (dev)))
			break;
	}
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
IEEE488write (dev_t dev, struct uio *uio, int flag)
{
	struct iovec *iov = uio->uio_iov;
	int ndone, nleft, ncopy;
	int err = 0;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	sc->uio = uio;
	nleft = DMA_BUFSIZE;
	ndone = 0;
	while (uio->uio_resid) {
		if (iov->iov_len > nleft)
			ncopy = nleft;
		else
			ncopy = iov->iov_len;
		if (copyin (iov->iov_base, sc->dmaBuffer + ndone, ncopy)) {
			err = EFAULT;
			break;
		}
		ndone += ncopy;
		nleft -= ncopy;
		iov->iov_base += ncopy;
		iov->iov_len -= ncopy;
		if (iov->iov_len == 0) {
			uio->uio_iovcnt--;
			iov++;
		}
		if ((nleft == 0) || (uio->uio_iovcnt == 0)) {
			err = [sc IEEE488ExecuteCommand:CMD_WRITE forDevice:minor(dev)];
			if (err)
				break;
			ndone = 0;
			nleft = DMA_BUFSIZE;
		}
	}
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
IEEE488ioctl (dev_t dev, int cmd, caddr_t data, int flag)
{
	int err;

	if (cmd == IEEE488_END) {
		*(int *)data = ((sc->readFinished & (1 << minor (dev))) != 0);
		return 0;
	}
	[sc->commandLock lockWhen:COMMAND_IDLE];
	sc->ioctlCmd = cmd;
	sc->ioctlData = data;
	err = [sc IEEE488ExecuteCommand:CMD_IOCTL forDevice:minor(dev)];
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}
